// This file was generated by ecm_setup_version(): DO NOT EDIT!

#ifndef KDBINDINGS_VERSION_H
#define KDBINDINGS_VERSION_H

#define KDBINDINGS_VERSION_STRING "1.0.95"
#define KDBINDINGS_VERSION_MAJOR 1
#define KDBINDINGS_VERSION_MINOR 0
#define KDBINDINGS_VERSION_PATCH 95
#define KDBINDINGS_VERSION KDBINDINGS_VERSION_CHECK(KDBINDINGS_VERSION_MAJOR, KDBINDINGS_VERSION_MINOR, KDBINDINGS_VERSION_PATCH)

/*
   for example: KDBINDINGS_VERSION >= KDBINDINGS_VERSION_CHECK(1, 2, 2))
*/
#define KDBINDINGS_VERSION_CHECK(major, minor, patch) ((major<<16)|(minor<<8)|(patch))

#endif
